#include "interpreter.h"

interpreter::interpreter (TextEditor* outputWindow)
: Thread(T("Interpret")),
  log (outputWindow)
{
    FileInputStream* sourceFile= new FileInputStream(File ("prueba.cod"));
    code = new CodeDocument();
    code->loadFromStream((InputStream&)*sourceFile);
    log->setText(T("Number of lines: ")+String(code->getNumLines())+T("\n"));
    nLines = code->getNumLines();
    varCounter=0;
    clearVarTable();
    interpretState=_STOPED_;
}

interpreter::~interpreter()
{

}


void interpreter::run(){
	interpretState=_RUNNING_;

	ifIndex=0;				// Numero de ifs anidados.
	nTags=0;				// Numero de etiquetas definidas.
	nJumps=0;				// Numero de saltos a subrutinas ejecutados y no retornados.

	loopIndex=0;
	programCounter=0;		// Contador de programa.
	waitState=false;
	log->setCaretPosition(log->getText().length());

	printf ("HILOS: %d\n",Thread::getNumRunningThreads());
	fflush(stdout);

	//Detectamos las etiquetas dispersas por el cdigo y les asignamos su numero de linea.
	for (int i=0;i<nLines;i++){
		String line = code->getLine(i).trim().toLowerCase();

		if (line[line.length()-1]==':')
		{
			String tagName = line.substring(0,line.length()-1);
			for (unsigned int j=0;j<nTags;j++){
				if (tagList[j].tagName == tagName){
					printf("ERROR! la etiqueta %s esta repetida.",tagName.toCString());
					fflush(stdout);
					return;
				}
			}

			tagList[nTags].tagName = tagName;
			tagList[nTags].lineNumber = i;
			nTags++;
		}
	}
	log->insertTextAtCaret("RUN...\n");
	log->getParentComponent()->repaint();

	// Ejecutamos linea a linea todo el codigo.
	do{
		runCode();
	}while (interpretState==_RUNNING_ && programCounter<nLines && !threadShouldExit());

	log->insertTextAtCaret("END!\n");
	interpretState=_STOPED_;
}

void interpreter::runCode(){

	int trueConditions=0;	// And de todas las condiciones de los ifs anidados.
	String line = code->getLine(programCounter).trim();

	if(line[0]!='#' && line.length()>0)
	{
		String command = line.upToFirstOccurrenceOf("[", false,true).trim().toLowerCase();
		String params = line.fromFirstOccurrenceOf("[", false,true).upToFirstOccurrenceOf("]", false,true);
		if (command!="print"){
			params = params.trim().toLowerCase();
		}

		//log->insertTextAtCaret(T("COMMAND: <")+command+">\n");
		//log->insertTextAtCaret(T("PARAM: ")+params+"\n");

		if (ifIndex>0){
			trueConditions=1;
			for (unsigned int i=0;i<ifIndex;i++){
				if (ifCondition[i]==false){
					trueConditions = 0;
				}
			}

			if (command=="end_if"){
				if (ifIndex>0){
					ifIndex--;
				}
			}
			else{
				if (command=="else"){
					ifCondition[ifIndex-1] = !ifCondition[ifIndex-1];
				}
				else{
					bool ifCmd=false;
					if (command=="if"){
						ifCmd=true;
					}

					if (trueConditions || ifCmd==true){
						decodeCommands(command,params);
					}
				}
			}
		}
		else{
			decodeCommands(command,params);
		}
	}

	//if (waitState==false){
		programCounter++;
	//}
	//else{
	//	if (Time::getMillisecondCounter()>waitEnd){
	//		log->insertTextAtCaret(String(Time::getMillisecondCounter())+"\n");
	//		waitState = false;
	//	}
	//}

}

void interpreter::decodeCommands(String command, String params){


	if (command=="if"){
		ifCondition[ifIndex] = valueCondition(params);
		ifIndex++;
	}

	if (command=="goto"){
		bool saltoEfectuado=false;

		for (unsigned int i=0; i<nTags;i++){
			if (params == tagList[i].tagName){
				programCounter = tagList[i].lineNumber;
				saltoEfectuado=true;
			}
		}

		if (saltoEfectuado==false){
			printf("ERROR! Etiqueta '%s' no identificada.\n",params.toCString());
		}
	}

	if (command=="gosub"){
		bool saltoEfectuado=false;

		for (unsigned int i=0; i<nTags;i++){
			if (params == tagList[i].tagName){
				if (nJumps<_MAX_JUMPS_){
					jumpList[nJumps++] = programCounter;
					programCounter = tagList[i].lineNumber;
					saltoEfectuado=true;
				}
				else{
					printf("ERROR! Te has pasado realizando saltos.\n");
				}
			}
		}

		if (saltoEfectuado==false){
			printf("ERROR LINEA %d! Etiqueta '%s' no identificada.\n", programCounter,params.toCString());
		}
	}

	if (command=="return"){
		if (nJumps>0){
			programCounter = jumpList[nJumps-1];
			nJumps--;
		}
		else{
			printf("ERROR. No se puede retornar.\n");
		}
	}

	if (command=="end"){
		programCounter = nLines;
	}

	if (command=="clear"){
		clearVarTable();
	}

	if (command=="print"){
		int varValue = getVarValue(params);
		if (varValue!=-9999){
			log->insertTextAtCaret(T(" ") + String(varValue) + T(" "));
		}
		else{
			log->insertTextAtCaret(params.upToFirstOccurrenceOf("/",false,true));
			if (params[params.length()-1]=='/'){
				log->insertTextAtCaret("\n");
			}
		}
	}

	if (command=="assign"){
		String varName = params.upToFirstOccurrenceOf(",",false,true).trim().toLowerCase();
		String varValue = params.fromFirstOccurrenceOf(",",false,true).trim().toLowerCase();

		if (varName.length()>0 && varValue.length()>0){
			assign(varName,varValue);
			//log->insertTextAtCaret("AADIENDO VARIABLE " + varName + "=" + varValue +"\n");
		}
	}

	if (command=="timestamp"){
		Time currentTime;
		log->insertTextAtCaret(T("[")+currentTime.getCurrentTime().toString(true,true,true,true)+ "]\n");
	}

	if (command=="inc"){
		String varName = params.upToFirstOccurrenceOf(",",false,true).trim().toLowerCase();
		int varValue = getVarValue(params);
		if (varValue!=-9999){
			varValue++;
			assign(varName,String(varValue));
		}else{
			//Variable no encontrada.
		}
	}

	if (command=="wait"){
		waitState = true;
		waitBegin = Time::getMillisecondCounter();
		waitEnd = waitBegin + params.getIntValue()*1000;
	}

	if (command=="loop"){
		// TODO. Hacer que acepte valores de variables adems de valores literales.
		loopJump[loopIndex]=programCounter;
		loopCnt[loopIndex] = params.getIntValue();
		loopIndex++;
	}

	if (command=="end_loop"){
		if (loopCnt[loopIndex-1]>1){
			loopCnt[loopIndex-1]--;
			programCounter = loopJump[loopIndex-1];
		}
		else{
			if (loopIndex>0){
				loopIndex=loopIndex-1;
			}
//					else{
//						log->insertTextAtCaret(T("ERROR LINE: ") + String(programCounter) + "... INCORRECT END OF LOOP.\n");
//					}
		}
	}
}

bool interpreter::valueCondition(String condition){

	String firstParam,secondParam;
	char logicOperator=' ';

	bool equal=false;
	int operatorA;
	int operatorB;

	//Search Operation
	String operations = "><!=";
	for (unsigned int i=0;i<operations.length() && logicOperator==' ';i++){
		firstParam = condition.upToFirstOccurrenceOf(String(&operations[i],1),false,true).trim();
		secondParam = condition.fromFirstOccurrenceOf(String(&operations[i],1), false,true).upToFirstOccurrenceOf  ("]", false,true).trim();

		if (secondParam[0]=='='){
			equal=true;
			secondParam = secondParam.fromFirstOccurrenceOf("=",false,true).trim();
		}

		if (firstParam.length()>0 && firstParam.length()<condition.length() &&
			secondParam.length()>0 && secondParam.length()<condition.length()){
				logicOperator = operations[i];
		}
	}

	//log->insertTextAtCaret(T("OPERATION: ") + firstParam + logicOperator + secondParam + "\n");

	operatorA = decodeParam(firstParam);
	operatorB = decodeParam(secondParam);

	//Process operation
	switch (logicOperator){
	case '>':
		if (equal==true){
			return (operatorA>=operatorB);
		}
		else{
			return (operatorA>operatorB);
		}
		break;

	case '<':
		if (equal==true){
			return (operatorA<=operatorB);
		}
		else{
			return (operatorA<operatorB);
		}
		break;

	case '!':
		return (operatorA!=operatorB); //Soporta operador '!=' y '!' a secas.
		break;

	case '=':
		return (operatorA==operatorB); //Soporta indistintamente '=' y '=='
		break;

	default:
		log->insertTextAtCaret("ERROR! CONDICION NO RECONOCIDA!\n");
		break;
	}

	return false;
}

int interpreter::decodeParam(String param){

	int operatorValue=0;
	bool operatorValueIsHexadecimal=false;

	//Detect numeric constants and variables.
	if (param[0]=='0' && param[1]=='x'){
		param = param.fromFirstOccurrenceOf("0x", false,true);
		operatorValueIsHexadecimal=true;
	}

	if (param.containsOnly("0123456789") && operatorValueIsHexadecimal==false){
		operatorValue = param.getIntValue();
		//log->insertTextAtCaret(T("OPERATOR = ")+String(operatorValue)+"\n");
	}
	else{
		if (param.containsOnly("0123456789ABCDEF") && operatorValueIsHexadecimal==true){
			operatorValue = param.getHexValue32();
			//log->insertTextAtCaret(T("OPERATOR (HEX) = ")+String(operatorValue)+"\n");
		}
		else{
			//Detect if var is defined and get its value.
			operatorValue=getVarValue(param);
			if (operatorValue==-9999){
				log->insertTextAtCaret("ERROR! VARIABLE '" + param + "' NOT FOUND!\n");
			}
		}
	}

	return operatorValue;
}

void interpreter::clearVarTable(){
	for (unsigned int i=0; i<_VAR_TABLE_SIZE_;i++){
		varTable[i].Name="";
		varTable[i].Value="";
	}
}

int interpreter::getVarValue(String varName){
	for (unsigned int i=0;i<varCounter;i++){
		if (varTable[i].Name.toLowerCase()==varName.toLowerCase()){
			return varTable[i].Value.getIntValue();
		}
	}

	return -9999;
}

int interpreter::assign(String varName, String value){

	bool flag=false;

	for (unsigned int i=0;i<_VAR_TABLE_SIZE_ && flag==false;i++){
		if (varTable[i].Name==varName){
			varTable[i].Value=value;
			flag=true;
		}
	}

	if (flag==false && varCounter<_VAR_TABLE_SIZE_){
		varTable[varCounter].Name=varName;
		varTable[varCounter].Value=value;
		varCounter++;
		flag=true;
	}

	if (flag==false){
		return -1;
	}

	return 0;
}
